{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE RankNTypes #-}

-- |
-- Module      : Data.ZSDD.Internal
-- Copyright   : (c) Eddie Jones 2020
-- License     : BSD-3
-- Maintainer  : eddiejones2108@gmail.com
-- Stability   : experimental
--
-- = WARNING
--
-- This module is __internal__ and may be frequently changed.
module Data.ZSDD.Internal
  ( -- * Diagrams
    Diagram (..),
    Ops (..),
    runDiagram,

    -- * Internal Nodes
    ID,
    Node (..),
    withNode,
    insertNode,
    lookupNode,

    -- * Propositions
    Prop (..),
    Invertible,
    makeProp,

    -- ** Construction
    -- $construction
    atomic,
    negate,
    assign,
    union,
    intersect,
    difference,

    -- ** Query
    isEmpty,
    isNotEmpty,
    count,

    -- ** Map
    mapAtom,
    bindAtom,
  )
where

import Control.Monad
import Control.Monad.State
import qualified Data.HashMap.Lazy as H
import Data.Hashable
import qualified Data.IntMap as I
import GHC.Generics (Generic)
import Prelude hiding (map, negate)

-- | A unique node identifier
type ID = Int

-- | A proposition parameterised by it's diagram and atom type
data Prop d a
  = Top
  | Bot
  | ID ID
  deriving (Eq, Ord, Generic)

instance Hashable (Prop d a)

-- | An internal node of the diagram graph
data Node d a = Node
  { atom :: a,
    lo :: Prop d a,
    hi :: Prop d a
  }
  deriving (Eq, Generic)

instance Hashable a => Hashable (Node d a)

-- | Atom types where every atom has an inverse.
-- This is only required by the difference operation
class Invertible a where
  neg :: a -> a

-- | The ambient decision diagram of propositions.
--
-- The type parameter d is the diagrams existential labelled and should not be instantiated by a concrete type
newtype Diagram d a m r = Diagram
  { getState ::
      StateT
        ( ID, -- Fresh node ID
          H.HashMap (Ops a) (Prop d a), -- Cached operations
          H.HashMap (Node d a) Int, -- node to ID
          I.IntMap (Node d a) -- ID to node
        )
        m
        r
  }
  deriving (Functor, Applicative, Monad)

-- | Operation DSL for cache
data Ops a
  = Atomic a
  | Union ID ID
  | Intersect ID ID
  | Difference ID ID
  | DifferenceT ID
  | Negate ID a
  | Assign ID a Bool
  deriving (Eq, Ord, Generic)

instance Hashable a => Hashable (Ops a)

-- | Extract information that does not depend on the digram
runDiagram :: Monad m => (forall d. Diagram d a m r) -> m r
runDiagram (Diagram f) = evalStateT f (0, H.empty, H.empty, I.empty)

-- Lookup operation in cache and run otherwise
memo :: (Ord a, Hashable a, Monad m) => Ops a -> Diagram d a m (Prop d a) -> Diagram d a m (Prop d a)
memo o k = Diagram $ do
  (_, mem, _, _) <- get
  case H.lookup o mem of
    Nothing -> do
      p <- getState k
      (i, mem', h, ns) <- get
      put (i, H.insert o p mem', h, ns)
      return p
    Just p -> return p

-- Get a fresh id
fresh :: Monad m => Diagram d a m ID
fresh = Diagram $ do
  (i, mem, h, ns) <- get
  put (i + 1, mem, h, ns)
  return i

-- | Extract internal node information from it's ID
withNode :: Monad m => ID -> (Node d a -> Diagram d a m r) -> Diagram d a m r
withNode i f = Diagram $ do
  (_, _, _, ns) <- get
  case I.lookup i ns of
    Nothing -> error ("Node not found!")
    Just n -> getState $ f n

-- | Insert a new node
insertNode :: (Eq a, Hashable a, Monad m) => ID -> Node d a -> Diagram d a m ()
insertNode i n = Diagram $ do
  (j, m, h, ns) <- get
  put (j, m, H.insert n i h, I.insert i n ns)

-- | Find the ID of an existing node
lookupNode :: (Eq a, Hashable a, Monad m) => Node d a -> Diagram d a m (Maybe ID)
lookupNode n = Diagram $ do
  (_, _, h, _) <- get
  return $ H.lookup n h

-- | Make a new proposition (if not already present) from a node
makeProp :: (Eq a, Hashable a, Monad m) => Node d a -> Diagram d a m (Prop d a)
makeProp n
  | hi n == Bot = return (lo n) -- Node elimination
  | otherwise = lookupNode n >>= \case
    Just i -> return (ID i)
    Nothing -> do
      i <- fresh
      insertNode i n
      return (ID i)

-- $construction
-- Non-trivial propositions are constructed within a diagram.
-- Although propositions use information from the diagram, they are independant;
-- each operation is applied to the supplied proposition leaving others untouched.

-- | Construct a proposition from an atom
atomic :: (Ord a, Hashable a, Monad m) => a -> Diagram d a m (Prop d a)
atomic a =
  memo (Atomic a) $
    makeProp Node {atom = a, lo = Bot, hi = Top}

-- | Instantiating an atom
assign :: (Ord a, Hashable a, Monad m) => Prop d a -> a -> Bool -> Diagram d a m (Prop d a)
assign Top _ _ = return Top
assign Bot _ _ = return Bot
assign (ID i) var b =
  memo (Assign i var b) $
    withNode
      i
      ( \n ->
          case compare (atom n) var of
            LT -> return Bot
            EQ -> return (hi n)
            GT -> do
              lo' <- assign (lo n) var b
              hi' <- assign (hi n) var b
              makeProp
                Node
                  { atom = atom n,
                    lo = lo',
                    hi = hi'
                  }
      )

-- | Negate every occurance of an atom
negate :: (Ord a, Hashable a, Monad m) => Prop d a -> a -> Diagram d a m (Prop d a)
negate Top _ = return Top
negate Bot _ = return Bot
negate p@(ID i) var =
  memo (Negate i var) $
    withNode
      i
      ( \n ->
          case compare (atom n) var of
            LT ->
              makeProp
                Node
                  { atom = var,
                    lo = Bot,
                    hi = p
                  }
            EQ ->
              makeProp
                Node
                  { atom = var,
                    lo = hi n,
                    hi = lo n
                  }
            GT -> do
              lo' <- negate (lo n) var
              hi' <- negate (hi n) var
              makeProp
                Node
                  { atom = atom n,
                    lo = lo',
                    hi = hi'
                  }
      )

-- | Union of two propositions
union :: (Ord a, Hashable a, Monad m) => Prop d a -> Prop d a -> Diagram d a m (Prop d a)
union Top _ = return Top
union _ Top = return Top
union Bot q = return q
union p Bot = return p
union p q | p == q = return p
union p@(ID i) (ID j) =
  memo (Union i j) $
    withNode
      i
      ( \n ->
          withNode j $ \m ->
            case compare (atom n) (atom m) of
              LT -> do
                lo' <- p `union` lo m
                makeProp
                  Node
                    { atom = atom m,
                      lo = lo',
                      hi = hi m
                    }
              EQ -> do
                lo' <- lo n `union` lo m
                hi' <- hi n `union` hi m
                makeProp
                  Node
                    { atom = atom n,
                      lo = lo',
                      hi = hi'
                    }
              GT -> do
                lo' <- lo n `union` p
                makeProp
                  Node
                    { atom = atom n,
                      lo = lo',
                      hi = hi n
                    }
      )

-- | Intersection of two propositions
intersect :: (Ord a, Hashable a, Monad m) => Prop d a -> Prop d a -> Diagram d a m (Prop d a)
intersect Top p = return p
intersect p Top = return p
intersect Bot _ = return Bot
intersect _ Bot = return Bot
intersect p q | p == q = return p
intersect p@(ID i) q@(ID j) =
  memo (Intersect i j) $
    withNode
      i
      ( \n ->
          withNode j $ \m ->
            case compare (atom n) (atom m) of
              LT -> lo n `intersect` q
              EQ -> do
                lo' <- lo n `intersect` lo m
                hi' <- hi n `intersect` hi m
                makeProp
                  Node
                    { atom = atom n,
                      lo = lo',
                      hi = hi'
                    }
              GT -> p `intersect` lo m
      )

-- | Difference between two propositions
difference :: (Ord a, Invertible a, Hashable a, Monad m) => Prop d a -> Prop d a -> Diagram d a m (Prop d a)
difference _ Top = return Bot
difference Bot _ = return Bot
difference p Bot = return p
difference p q | p == q = return p
difference Top (ID j) =
  memo (DifferenceT j) $
    withNode
      j
      ( \m ->
          makeProp
            Node
              { atom = neg $ atom m,
                lo = lo m,
                hi = hi m
              }
      )
difference p@(ID i) q@(ID j) =
  memo (Difference i j) $
    withNode
      i
      ( \n ->
          withNode j $ \m ->
            case compare (atom n) (atom m) of
              LT -> p `difference` lo m
              EQ -> do
                lo' <- lo n `difference` lo m
                hi' <- hi n `difference` hi m
                makeProp
                  Node
                    { atom = atom n,
                      lo = lo',
                      hi = hi'
                    }
              GT -> do
                lo' <- lo n `difference` q
                makeProp
                  Node
                    { atom = atom n,
                      lo = lo',
                      hi = hi n
                    }
      )

-- | Does the proposition have any models?
isEmpty, isNotEmpty :: Monad m => Prop d a -> Diagram d a m Bool
isEmpty Bot = return True
isEmpty Top = return False
isEmpty (ID i) = withNode i $ \n -> do
  p <- isEmpty (lo n)
  q <- isEmpty (hi n)
  return (p && q)
isNotEmpty Bot = return False
isNotEmpty Top = return True
isNotEmpty (ID i) = withNode i $ \n -> do
  p <- isNotEmpty (lo n)
  q <- isNotEmpty (hi n)
  return (p || q)

-- | Count the number of models
count :: Monad m => Prop d a -> Diagram d a m Int
count Bot = return 0
count Top = return 1
count (ID i) = withNode i $ \n -> do
  x <- count (lo n)
  y <- count (hi n)
  return (x + y)

-- | Map between atoms
mapAtom :: (Eq a, Hashable a, Monad m) => (a -> a) -> Prop d a -> Diagram d a m (Prop d a)
mapAtom _ Top = return Top
mapAtom _ Bot = return Bot
mapAtom f (ID i) = withNode i $ \n -> do
  lo' <- mapAtom f (lo n)
  hi' <- mapAtom f (hi n)
  makeProp
    Node
      { atom = f (atom n),
        lo = lo',
        hi = hi'
      }

-- Loop continuations to another part of the diagram
mapTerminal :: (Eq a, Hashable a, Monad m) => Prop d a -> Prop d a -> Prop d a -> Diagram d a m (Prop d a)
mapTerminal top _ Top = return top
mapTerminal _ bot Bot = return bot
mapTerminal top bot (ID i) = withNode i $ \n -> do
  lo' <- mapTerminal top bot (lo n)
  hi' <- mapTerminal top bot (hi n)
  makeProp
    Node
      { atom = atom n,
        lo = lo',
        hi = hi'
      }

-- | Instantiate an atom with a non-atomic proposition
bindAtom :: (Eq a, Hashable a, Monad m) => (a -> Diagram d a m (Prop d a)) -> Prop d a -> Diagram d a m (Prop d a)
bindAtom _ Top = return Top
bindAtom _ Bot = return Bot
bindAtom f (ID i) = withNode i $ \n -> do
  lo' <- bindAtom f (lo n)
  hi' <- bindAtom f (hi n)
  p <- f (atom n)
  mapTerminal lo' hi' p
